"""Define fixtures, constants, etc. available for all tests."""

import json
import re
from collections.abc import Generator
from time import time
from typing import Any, cast

import aiohttp
import pytest
from aresponses import ResponsesMockServer

from .common import TILE_CLIENT_UUID, TILE_EMAIL, TILE_USER_UUID, load_fixture


@pytest.fixture(name="authenticated_tile_api_server")
def authenticated_tile_api_server_fixture(
    create_client_response: dict[str, Any],
    create_session_response: dict[str, Any],
) -> Generator[ResponsesMockServer]:
    """Return a fixture that mocks an authenticated tile API server.

    Args:
        create_client_response: An API response payload
        create_session_response: An API response payload
    """
    client_pattern = re.compile(r"/api/v1/clients/.+")
    session_pattern = re.compile(r"/api/v1/clients/.+/sessions")

    server = ResponsesMockServer()
    server.add(
        "production.tile-api.com",
        client_pattern,
        "put",
        response=aiohttp.web_response.json_response(create_client_response, status=200),
    )
    server.add(
        "production.tile-api.com",
        session_pattern,
        "post",
        response=aiohttp.web_response.json_response(
            create_session_response, status=200
        ),
    )
    yield server


@pytest.fixture(name="create_client_response", scope="session")
def create_client_response_fixture() -> dict[str, Any]:
    """Return a fixture for a successful client creation payload."""
    return cast(dict[str, Any], json.loads(load_fixture("create_client_response.json")))


@pytest.fixture(name="create_session_response")
def create_session_response_fixture() -> dict[str, Any]:
    """Return a /clients/<UUID>/sessions response.

    Returns:
        An API response payload.
    """
    return {
        "version": 1,
        "revision": 1,
        "timestamp": "2018-06-19T23:04:24.672Z",
        "timestamp_ms": 1529449464672,
        "result_code": 0,
        "result": {
            "client_uuid": TILE_CLIENT_UUID,
            "user": {
                "user_uuid": TILE_USER_UUID,
                "full_name": None,
                "email": TILE_EMAIL,
                "beta_eligibility": False,
                "gift_recipient": True,
                "locale": "en-US",
                "email_shared": True,
                "image_url": None,
                "status": "ACTIVATED",
                "pw_exists": True,
                "linked_accounts": [],
                "registration_timestamp": 1482711582203,
                "last_modified_timestamp": 1529444807328,
            },
            "session_start_timestamp": int(time() * 1000),
            "session_expiration_timestamp": int(time() * 1000) + 1000,
            "changes": "EXISTING_ACCOUNT",
        },
    }


@pytest.fixture(name="expired_session_response")
def expired_session_response_fixture() -> dict[str, Any]:
    """Return a /clients/<UUID>/sessions response with an expired session.

    Returns:
        An API response payload.
    """
    return {
        "version": 1,
        "revision": 1,
        "timestamp": "2018-06-19T23:04:24.672Z",
        "timestamp_ms": 1529449464672,
        "result_code": 0,
        "result": {
            "client_uuid": TILE_CLIENT_UUID,
            "user": {
                "user_uuid": TILE_USER_UUID,
                "full_name": None,
                "email": TILE_EMAIL,
                "beta_eligibility": False,
                "gift_recipient": True,
                "locale": "en-US",
                "email_shared": True,
                "image_url": None,
                "status": "ACTIVATED",
                "pw_exists": True,
                "linked_accounts": [],
                "registration_timestamp": 1482711582203,
                "last_modified_timestamp": 1529444807328,
            },
            "session_start_timestamp": int(time() * 1000),
            "session_expiration_timestamp": int(time() * 1000) - 1000,
            "changes": "EXISTING_ACCOUNT",
        },
    }


@pytest.fixture(name="tile_details_missing_last_state_response", scope="session")
def tile_details_missing_last_state_response_fixture() -> dict[str, Any]:
    """Return a fixture for a Tile details response that's missing it's state info."""
    return cast(
        dict[str, Any],
        json.loads(load_fixture("tile_details_missing_last_state_response.json")),
    )


@pytest.fixture(name="tile_details_new_name_response_fixture")
def tile_details_new_name_response_fixture() -> dict[str, Any]:
    """Define a fixture for a Tile with a new name.

    Returns:
        An API response payload.
    """
    data = json.loads(load_fixture("tile_details_response.json"))
    data["result"]["name"] = "New Name"
    return cast(dict[str, Any], data)


@pytest.fixture(name="tile_details_response", scope="session")
def tile_details_response_fixture() -> dict[str, Any]:
    """Return a fixture for a Tile details response."""
    return cast(dict[str, Any], json.loads(load_fixture("tile_details_response.json")))


@pytest.fixture(name="tile_details_update_response")
def tile_details_update_response_fixture() -> dict[str, Any]:
    """Define a fixture for a Tile with updated coordinates.

    Returns:
        An API response payload.
    """
    data = json.loads(load_fixture("tile_details_response.json"))
    data["result"]["last_tile_state"]["latitude"] = 51.8943631
    data["result"]["last_tile_state"]["longitude"] = -0.4930538
    return cast(dict[str, Any], data)


@pytest.fixture(name="tile_history_response", scope="session")
def tile_history_response_fixture() -> dict[str, Any]:
    """Return a fixture for a Tile history response."""
    return cast(dict[str, Any], json.loads(load_fixture("tile_history_response.json")))


@pytest.fixture(name="tile_states_response", scope="session")
def tile_states_response_fixture() -> dict[str, Any]:
    """Return a fixture for a Tile states response."""
    return cast(dict[str, Any], json.loads(load_fixture("tile_states_response.json")))



"""Define an object to work directly with the API."""

from __future__ import annotations

import asyncio
from time import time
from typing import Any, cast
from uuid import uuid4

from aiohttp import ClientSession
from aiohttp.client_exceptions import ClientError

from .const import LOGGER
from .errors import InvalidAuthError, RequestError
from .tile import Tile

API_URL_SCAFFOLD = "https://production.tile-api.com/api/v1"

DEFAULT_API_VERSION = "1.0"
DEFAULT_APP_ID = "ios-tile-production"
DEFAULT_APP_VERSION = "2.89.1.4774"
DEFAULT_LOCALE = "en-US"
DEFAULT_TIMEOUT = 10
DEFAULT_USER_AGENT = "Tile/4774 CFNetwork/1312 Darwin/21.0.0"


class API:
    """Define the API management object."""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        email: str,
        password: str,
        session: ClientSession,
        *,
        client_uuid: str | None = None,
        locale: str = DEFAULT_LOCALE,
    ) -> None:
        """Initialize.

        Args:
            email: An email address for a Tile account.
            password: The account password.
            session: An optional aiohttp ClientSession.
            client_uuid: An optional UUID to identify this API object.
            locale: An optional locale.
        """
        self._client_established: bool = False
        self._email: str = email
        self._locale: str = locale
        self._password: str = password
        self._session: ClientSession = session
        self._session_expiry: int | None = None
        self.client_uuid: str = client_uuid if client_uuid else str(uuid4())
        self.user_uuid: str | None = None

    async def _async_request(
        self, method: str, endpoint: str, **kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        """Make an API request.

        Args:
            method: An HTTP method.
            endpoint: A relative API endpoint.
            **kwargs: Additional kwargs to send with the request.

        Returns:
            An API response payload.

        Raises:
            InvalidAuthError: Raised upon invalid credentials.
            RequestError: Raised upon an underlying HTTP error.
        """
        if self._session_expiry and self._session_expiry <= int(time() * 1000):
            await self.async_init()

        kwargs.setdefault("headers", {})
        kwargs["headers"]["User-Agent"] = DEFAULT_USER_AGENT
        kwargs["headers"]["tile_api_version"] = DEFAULT_API_VERSION
        kwargs["headers"]["tile_app_id"] = DEFAULT_APP_ID
        kwargs["headers"]["tile_app_version"] = DEFAULT_APP_VERSION
        kwargs["headers"]["tile_client_uuid"] = self.client_uuid

        try:
            async with self._session.request(
                method, f"{API_URL_SCAFFOLD}/{endpoint}", **kwargs
            ) as resp:
                resp.raise_for_status()
                data = await resp.json()
        except ClientError as err:
            if "401" in str(err):
                raise InvalidAuthError("Invalid credentials") from err
            raise RequestError(f"Error requesting data from {endpoint}: {err}") from err

        LOGGER.debug("Data received from /%s: %s", endpoint, data)

        return cast(dict[str, Any], data)

    async def async_get_tiles(self) -> dict[str, Tile]:
        """Get all active Tiles from the user's account.

        Returns:
            A dictionary of Tile UUIDs to Tile objects.
        """
        states = await self._async_request("get", "tiles/tile_states")

        details_tasks = {
            tile_uuid: self._async_request("get", f"tiles/{tile_uuid}")
            for tile_uuid in [tile["tile_id"] for tile in states["result"]]
        }

        results = await asyncio.gather(*details_tasks.values(), return_exceptions=True)

        data = {}
        for tile_uuid, result in zip(details_tasks, results):
            if isinstance(result, RequestError) and "412" in str(result):
                # Tile Labels will return an HTTP 412 because they don't have
                # additional details; we can safely ignore these errors and still
                # track the Tile (without additional details):
                continue
            if isinstance(result, BaseException):
                LOGGER.error("Error requesting details for %s: %s", tile_uuid, result)
                continue
            data[tile_uuid] = Tile(self._async_request, result)

        return data

    async def async_init(self) -> None:
        """Create a Tile session."""
        # Invalidate the existing session expiry datetime (if it exists) so that the
        # next few requests don't fail:
        self._session_expiry = None

        if not self._client_established:
            await self._async_request(
                "put",
                f"clients/{self.client_uuid}",
                data={
                    "app_id": DEFAULT_APP_ID,
                    "app_version": DEFAULT_APP_VERSION,
                    "locale": self._locale,
                },
            )
            self._client_established = True

        resp = await self._async_request(
            "post",
            f"clients/{self.client_uuid}/sessions",
            data={"email": self._email, "password": self._password},
        )

        if not self.user_uuid:
            self.user_uuid = resp["result"]["user"]["user_uuid"]
        self._session_expiry = resp["result"]["session_expiration_timestamp"]


async def async_login(
    email: str,
    password: str,
    session: ClientSession,
    *,
    client_uuid: str | None = None,
    locale: str = DEFAULT_LOCALE,
) -> API:
    """Return an authenticated client.

    Args:
        email: An email address for a Tile account.
        password: The account password.
        session: An optional aiohttp ClientSession.
        client_uuid: An optional UUID to identify this API object.
        locale: An optional locale.

    Returns:
        An authenticated API object.
    """
    api: API = API(email, password, session, client_uuid=client_uuid, locale=locale)
    await api.async_init()
    return api



"""Define package errors."""


class TileError(Exception):
    """Define a base error."""

    pass


class InvalidAuthError(TileError):
    """Define an error related to invalid authentication."""

    pass


class RequestError(TileError):
    """Define an error related to invalid requests."""

    pass


class SessionExpiredError(TileError):
    """Define an error for when a Tile app session expires."""

    pass



"""Define common test utilities."""

import os

TILE_CLIENT_UUID = "2cc56adc-b96a-4293-9b94-eda716e0aa17"
TILE_EMAIL = "user@email.com"
TILE_PASSWORD = "12345"  # noqa: S105
TILE_TILE_NAME = "Wallet"
TILE_TILE_UUID = "19264d2dffdbca32"
TILE_USER_UUID = "fd0c10a5-d0f7-4619-9bce-5b2cb7a6754b"


def load_fixture(filename: str) -> str:
    """Load a fixture.

    Args:
        filename: The filename of the fixtures/ file to load.

    Returns:
        A string containing the contents of the file.
    """
    path = os.path.join(os.path.dirname(__file__), "fixtures", filename)
    with open(path, encoding="utf-8") as fptr:
        return fptr.read()



"""Define tests for the client object."""

import re
from time import time
from typing import Any

import aiohttp
import pytest
from aresponses import ResponsesMockServer

from pytile import async_login
from pytile.errors import InvalidAuthError, RequestError

from .common import (
    TILE_CLIENT_UUID,
    TILE_EMAIL,
    TILE_PASSWORD,
    TILE_TILE_UUID,
    TILE_USER_UUID,
)


@pytest.mark.asyncio
async def test_bad_endpoint(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
) -> None:
    """Test that an exception is raised on a bad endpoint.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
    """
    async with authenticated_tile_api_server:
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            "/api/v1/bad_endpoint",
            "get",
            aresponses.Response(text="", status=404),
        )

        with pytest.raises(RequestError):
            async with aiohttp.ClientSession() as session:
                api = await async_login(
                    TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
                )
                await api._async_request(  # pylint: disable=protected-access
                    "get", "bad_endpoint"
                )

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_expired_session(  # pylint: disable=too-many-arguments,too-many-positional-arguments
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
    create_client_response: dict[str, Any],
    create_session_response: dict[str, Any],
    tile_details_response: dict[str, Any],
    tile_states_response: dict[str, Any],
) -> None:
    """Test that an expired session is recreated automatically.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
        create_client_response: An API response payload.
        create_session_response: An API response payload.
        tile_details_response: An API response payload.
        tile_states_response: An API response payload.
    """
    async with authenticated_tile_api_server:
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/clients/{TILE_CLIENT_UUID}",
            "put",
            response=aiohttp.web_response.json_response(
                create_client_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/clients/{TILE_CLIENT_UUID}/sessions",
            "post",
            response=aiohttp.web_response.json_response(
                create_session_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            "/api/v1/tiles/tile_states",
            "get",
            response=aiohttp.web_response.json_response(
                tile_states_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aiohttp.web_response.json_response(
                tile_details_response, status=200
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = await async_login(
                TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
            )

            # Simulate an expired session:
            api._session_expiry = (  # pylint: disable=protected-access
                int(time() * 1000) - 1000000
            )
            await api.async_get_tiles()

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_invalid_auth(aresponses: ResponsesMockServer) -> None:
    """Test initializing a client with a Tile session."""
    client_pattern = re.compile(r"/api/v1/clients/.+")

    aresponses.add(
        "production.tile-api.com",
        client_pattern,
        "put",
        aresponses.Response(
            text="",
            status=401,
            headers={"Content-Type": "application/json"},
        ),
    )

    async with aiohttp.ClientSession() as session:
        with pytest.raises(InvalidAuthError):
            await async_login(TILE_EMAIL, TILE_PASSWORD, session)

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_login(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
) -> None:
    """Test initializing a client with a Tile session.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
    """
    async with authenticated_tile_api_server, aiohttp.ClientSession() as session:
        api = await async_login(TILE_EMAIL, TILE_PASSWORD, session)
        assert isinstance(api.client_uuid, str)
        assert api.client_uuid != TILE_CLIENT_UUID
        assert api.user_uuid == TILE_USER_UUID

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_login_existing(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
) -> None:
    """Test the creation of a client with an existing client UUID.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
    """
    async with authenticated_tile_api_server, aiohttp.ClientSession() as session:
        api = await async_login(
            TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
        )
        assert api.client_uuid == TILE_CLIENT_UUID

    aresponses.assert_plan_strictly_followed()



"""Define a Tile object."""

from __future__ import annotations

from collections.abc import Awaitable, Callable
from datetime import datetime, timezone
from typing import Any, cast

from .const import LOGGER


class Tile:
    """Define a Tile."""

    def __init__(
        self,
        async_request: Callable[..., Awaitable[dict[str, Any]]],
        tile_data: dict[str, Any],
    ) -> None:
        """Initialize.

        Args:
            async_request: The request method from the Client object.
            tile_data: A dictionary of Tile data.
        """
        self._async_request = async_request
        self._tile_data = tile_data

        self._last_timestamp: datetime | None = None
        self._lost_timestamp: datetime | None = None
        self._save_timestamps(tile_data)

    def __str__(self) -> str:
        """Return the string representation of the Tile.

        Returns:
            A string representation.
        """
        return f"<Tile uuid={self.uuid} name={self.name}>"

    @property
    def accuracy(self) -> float | None:
        """Return the accuracy of the last measurement.

        Returns:
            The accuracy (if it exists).
        """
        if (last_state := self._tile_data["result"].get("last_tile_state")) is None:
            return None
        return cast(float, last_state["h_accuracy"])

    @property
    def altitude(self) -> float | None:
        """Return the last detected altitude.

        Returns:
            The altitude (if it exists).
        """
        if (last_state := self._tile_data["result"].get("last_tile_state")) is None:
            return None
        return cast(float, last_state["altitude"])

    @property
    def archetype(self) -> str:
        """Return the archetype.

        Returns:
            The archetype.
        """
        return cast(str, self._tile_data["result"]["archetype"])

    @property
    def dead(self) -> bool:
        """Return whether the Tile is dead.

        Returns:
            The dead status.
        """
        return cast(bool, self._tile_data["result"]["is_dead"])

    @property
    def firmware_version(self) -> str:
        """Return the firmware version.

        Returns:
            The firmware version.
        """
        return cast(str, self._tile_data["result"]["firmware_version"])

    @property
    def hardware_version(self) -> str:
        """Return the hardware version.

        Returns:
            The hardware version.
        """
        return cast(str, self._tile_data["result"]["hw_version"])

    @property
    def kind(self) -> str:
        """Return the type of Tile.

        Returns:
            The type.
        """
        return cast(str, self._tile_data["result"]["tile_type"])

    @property
    def last_timestamp(self) -> datetime | None:
        """Return the timestamp of the last location measurement.

        Returns:
            The timestamp (if it exists).
        """
        return self._last_timestamp

    @property
    def latitude(self) -> float | None:
        """Return the last detected latitude.

        Returns:
            The latitude (if it exists).
        """
        if (last_state := self._tile_data["result"].get("last_tile_state")) is None:
            return None
        return cast(float, last_state["latitude"])

    @property
    def longitude(self) -> float | None:
        """Return the last detected longitude.

        Returns:
            The longitude (if it exists).
        """
        if (last_state := self._tile_data["result"].get("last_tile_state")) is None:
            return None
        return cast(float, last_state["longitude"])

    @property
    def lost(self) -> bool:
        """Return whether the Tile is lost.

        Since the Tile API can sometimes fail to return last_tile_state data, if it's
        missing here, we return True (indicating the Tile *is* lost).

        Returns:
            The lost status.
        """
        if (last_state := self._tile_data["result"].get("last_tile_state")) is None:
            return True
        return cast(bool, last_state["is_lost"])

    @property
    def lost_timestamp(self) -> datetime | None:
        """Return the timestamp when the Tile was last in a "lost" state.

        Returns:
            The timestamp (if it exists).
        """
        return self._lost_timestamp

    @property
    def name(self) -> str:
        """Return the name.

        Returns:
            The name.
        """
        return cast(str, self._tile_data["result"]["name"])

    @property
    def ring_state(self) -> str | None:
        """Return the ring state.

        Returns:
            The ring state (if it exists).
        """
        if (last_state := self._tile_data["result"].get("last_tile_state")) is None:
            return None
        return cast(str, last_state["ring_state"])

    @property
    def uuid(self) -> str:
        """Return the UUID.

        Returns:
            The UUID.
        """
        return cast(str, self._tile_data["result"]["tile_uuid"])

    @property
    def visible(self) -> bool:
        """Return whether the Tile is visible.

        Returns:
            The visibility.
        """
        return cast(bool, self._tile_data["result"]["visible"])

    @property
    def voip_state(self) -> str | None:
        """Return the VoIP state.

        Returns:
            The VoIP state (if it exists).
        """
        if (last_state := self._tile_data["result"].get("last_tile_state")) is None:
            return None
        return cast(str, last_state["voip_state"])

    def _save_timestamps(self, tile_data: dict[str, Any]) -> None:
        """Save UTC timestamps from a Tile data set.

        Args:
            tile_data: A dictionary of Tile data.
        """
        if (last_state := tile_data["result"].get("last_tile_state")) is None:
            LOGGER.warning("Missing last_tile_state; can't report location info")
            self._last_timestamp = None
            self._lost_timestamp = None
            return

        self._last_timestamp = datetime.fromtimestamp(
            last_state["timestamp"] / 1000, tz=timezone.utc
        ).replace(tzinfo=None)
        self._lost_timestamp = datetime.fromtimestamp(
            last_state["lost_timestamp"] / 1000, tz=timezone.utc
        ).replace(tzinfo=None)

    def as_dict(self) -> dict[str, Any]:
        """Return dictionary version of this Tile.

        Returns:
            A dictionary representation of the Tile.
        """
        return {
            "accuracy": self.accuracy,
            "altitude": self.altitude,
            "archetype": self.archetype,
            "dead": self.dead,
            "firmware_version": self.firmware_version,
            "hardware_version": self.hardware_version,
            "kind": self.kind,
            "last_timestamp": self.last_timestamp,
            "latitude": self.latitude,
            "longitude": self.longitude,
            "lost": self.lost,
            "lost_timestamp": self.lost_timestamp,
            "name": self.name,
            "ring_state": self.ring_state,
            "uuid": self.uuid,
            "visible": self.visible,
            "voip_state": self.voip_state,
        }

    async def async_history(
        self, start_datetime: datetime, end_datetime: datetime
    ) -> dict[str, Any]:
        """Get the latest measurements from the Tile.

        Returns:
            A dictionary containing the requested history.
        """
        return await self._async_request(
            "get",
            f"tiles/location/history/{self.uuid}",
            params={
                "aggregation": "False",
                "end_ts": round(end_datetime.timestamp() * 1000),
                "start_ts": round(start_datetime.timestamp() * 1000),
            },
        )

    async def async_update(self) -> None:
        """Get the latest measurements from the Tile."""
        data = await self._async_request("get", f"tiles/{self.uuid}")
        self._save_timestamps(data)
        self._tile_data = data



"""Define tests for the client object."""

import logging
from datetime import datetime
from typing import Any
from unittest.mock import Mock

import aiohttp
import pytest
from aresponses import ResponsesMockServer

from pytile import async_login
from pytile.tile import Tile

from .common import (
    TILE_CLIENT_UUID,
    TILE_EMAIL,
    TILE_PASSWORD,
    TILE_TILE_NAME,
    TILE_TILE_UUID,
)


@pytest.mark.asyncio
async def test_get_history(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
    tile_details_response: dict[str, Any],
    tile_history_response: dict[str, Any],
    tile_states_response: dict[str, Any],
) -> None:
    """Test getting all Tiles associated with an account.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
        tile_details_response: An API response payload.
        tile_history_response: An API response payload.
        tile_states_response: An API response payload.
    """
    async with authenticated_tile_api_server:
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            "/api/v1/tiles/tile_states",
            "get",
            response=aiohttp.web_response.json_response(
                tile_states_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aiohttp.web_response.json_response(
                tile_details_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/location/history/{TILE_TILE_UUID}",
            "get",
            response=aiohttp.web_response.json_response(
                tile_history_response, status=200
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = await async_login(
                TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
            )
            tiles = await api.async_get_tiles()
            tile = tiles[TILE_TILE_UUID]

            start_datetime = datetime(2023, 1, 1, 0, 0, 0)
            end_datetime = datetime(2023, 1, 31, 0, 0, 0)
            history = await tile.async_history(start_datetime, end_datetime)
            assert history == tile_history_response

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_get_tiles(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
    tile_details_response: dict[str, Any],
    tile_states_response: dict[str, Any],
) -> None:
    """Test getting all Tiles associated with an account.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
        tile_details_response: An API response payload.
        tile_states_response: An API response payload.
    """
    async with authenticated_tile_api_server:
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            "/api/v1/tiles/tile_states",
            "get",
            response=aiohttp.web_response.json_response(
                tile_states_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aiohttp.web_response.json_response(
                tile_details_response, status=200
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = await async_login(
                TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
            )
            tiles = await api.async_get_tiles()
            assert len(tiles) == 1

            tile = tiles[TILE_TILE_UUID]
            assert isinstance(tile, Tile)
            assert str(tile) == f"<Tile uuid={TILE_TILE_UUID} name={TILE_TILE_NAME}>"
            assert tile.accuracy == 13.496111
            assert tile.altitude == 0.4076319168123
            assert tile.archetype == "WALLET"
            assert not tile.dead
            assert not tile.lost
            assert tile.firmware_version == "01.12.14.0"
            assert tile.hardware_version == "02.09"
            assert tile.kind == "TILE"
            assert tile.last_timestamp == datetime(2020, 8, 12, 17, 55, 26)
            assert tile.latitude == 51.528308
            assert tile.longitude == -0.3817765
            assert tile.lost_timestamp == datetime(1969, 12, 31, 23, 59, 59, 999000)
            assert tile.name == TILE_TILE_NAME
            assert tile.ring_state == "STOPPED"
            assert tile.uuid == TILE_TILE_UUID
            assert tile.visible
            assert tile.voip_state == "OFFLINE"

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_get_tiles_http_error(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
    caplog: Mock,
    tile_states_response: dict[str, Any],
) -> None:
    """Test getting all Tiles associated with an account.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
        caplog: A mocked logging utility.
        tile_states_response: An API response payload.
    """
    caplog.set_level(logging.INFO)

    async with authenticated_tile_api_server:
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            "/api/v1/tiles/tile_states",
            "get",
            response=aiohttp.web_response.json_response(
                tile_states_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aresponses.Response(text=None, status=500),
        )

        async with aiohttp.ClientSession() as session:
            api = await async_login(
                TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
            )
            tiles = await api.async_get_tiles()
            assert len(tiles) == 0
            assert any("Error requesting details" in e.message for e in caplog.records)

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_missing_last_tile_state(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
    tile_details_missing_last_state_response: dict[str, Any],
    tile_states_response: dict[str, Any],
) -> None:
    """Test that a missing last_tile_state is handled correctly.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
        tile_details_missing_last_state_response: An API response payload.
        tile_states_response: An API response payload.
    """
    async with authenticated_tile_api_server:
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            "/api/v1/tiles/tile_states",
            "get",
            response=aiohttp.web_response.json_response(
                tile_states_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aiohttp.web_response.json_response(
                tile_details_missing_last_state_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aiohttp.web_response.json_response(
                tile_details_missing_last_state_response, status=200
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = await async_login(
                TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
            )
            tiles = await api.async_get_tiles()
            tile = tiles[TILE_TILE_UUID]
            assert not tile.accuracy
            assert not tile.altitude
            assert not tile.last_timestamp
            assert not tile.latitude
            assert not tile.longitude
            assert tile.lost
            assert not tile.lost_timestamp
            assert not tile.ring_state
            assert not tile.voip_state

            await tile.async_update()
            assert not tile.latitude
            assert not tile.longitude

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_tile_as_dict(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
    tile_details_response: dict[str, Any],
    tile_states_response: dict[str, Any],
) -> None:
    """Test dumping a Tile as a dictionary.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
        tile_details_response: An API response payload.
        tile_states_response: An API response payload.
    """
    async with authenticated_tile_api_server:
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            "/api/v1/tiles/tile_states",
            "get",
            response=aiohttp.web_response.json_response(
                tile_states_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aiohttp.web_response.json_response(
                tile_details_response, status=200
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = await async_login(
                TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
            )
            tiles = await api.async_get_tiles()
            assert len(tiles) == 1
            tile = tiles[TILE_TILE_UUID]
            assert tile.as_dict() == {
                "accuracy": 13.496111,
                "altitude": 0.4076319168123,
                "archetype": "WALLET",
                "dead": False,
                "firmware_version": "01.12.14.0",
                "hardware_version": "02.09",
                "kind": "TILE",
                "last_timestamp": datetime(2020, 8, 12, 17, 55, 26),
                "latitude": 51.528308,
                "longitude": -0.3817765,
                "lost": False,
                "lost_timestamp": datetime(1969, 12, 31, 23, 59, 59, 999000),
                "name": "Wallet",
                "ring_state": "STOPPED",
                "uuid": "19264d2dffdbca32",
                "visible": True,
                "voip_state": "OFFLINE",
            }

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_tile_label(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
    tile_states_response: dict[str, Any],
) -> None:
    """Test a Tile label (which is ignored by this library).

    This is a sanity check that a Tile Label, which returns an HTTP 412 when passed to
    /api/v1/tiles/{TILE_UUID}, is handled correctly.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
        tile_states_response: An API response payload.
    """
    async with authenticated_tile_api_server:
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            "/api/v1/tiles/tile_states",
            "get",
            response=aiohttp.web_response.json_response(
                tile_states_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aresponses.Response(text="", status=412),
        )

        async with aiohttp.ClientSession() as session:
            api = await async_login(
                TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
            )
            tiles = await api.async_get_tiles()
            assert len(tiles) == 0

    aresponses.assert_plan_strictly_followed()


@pytest.mark.asyncio
async def test_tile_update(
    aresponses: ResponsesMockServer,
    authenticated_tile_api_server: ResponsesMockServer,
    tile_details_response: dict[str, Any],
    tile_details_update_response: dict[str, Any],
    tile_states_response: dict[str, Any],
) -> None:
    """Test updating a Tile's status.

    Args:
        aresponses: An aresponses server.
        authenticated_tile_api_server: A mock Tile API server connection.
        tile_details_response: An API response payload.
        tile_details_update_response: An API response payload.
        tile_states_response: An API response payload.
    """
    async with authenticated_tile_api_server:
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            "/api/v1/tiles/tile_states",
            "get",
            response=aiohttp.web_response.json_response(
                tile_states_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aiohttp.web_response.json_response(
                tile_details_response, status=200
            ),
        )
        authenticated_tile_api_server.add(
            "production.tile-api.com",
            f"/api/v1/tiles/{TILE_TILE_UUID}",
            "get",
            response=aiohttp.web_response.json_response(
                tile_details_update_response, status=200
            ),
        )

        async with aiohttp.ClientSession() as session:
            api = await async_login(
                TILE_EMAIL, TILE_PASSWORD, session, client_uuid=TILE_CLIENT_UUID
            )
            tiles = await api.async_get_tiles()
            tile = tiles[TILE_TILE_UUID]
            assert tile.latitude == 51.528308
            assert tile.longitude == -0.3817765

            await tile.async_update()
            assert tile.latitude == 51.8943631
            assert tile.longitude == -0.4930538

    aresponses.assert_plan_strictly_followed()



"""Run an example script to quickly test."""

import asyncio
import logging

from aiohttp import ClientSession

from pytile import async_login
from pytile.errors import TileError

_LOGGER = logging.getLogger(__name__)

TILE_EMAIL = "<EMAIL>"
TILE_PASSWORD = "<PASSWORD>"  # noqa: S105


async def main() -> None:
    """Run."""
    logging.basicConfig(level=logging.INFO)
    async with ClientSession() as session:
        try:
            api = await async_login(TILE_EMAIL, TILE_PASSWORD, session)

            tiles = await api.async_get_tiles()
            _LOGGER.info("Tile Count: %s", len(tiles))
            for tile in tiles.values():
                _LOGGER.info("UUID: %s", tile.uuid)
                _LOGGER.info("Name: %s", tile.name)
                _LOGGER.info("Type: %s", tile.kind)
                _LOGGER.info("Latitude: %s", tile.latitude)
                _LOGGER.info("Longitude: %s", tile.longitude)
                _LOGGER.info("Last Timestamp: %s", tile.last_timestamp)
        except TileError as err:
            _LOGGER.info(err)


asyncio.run(main())
